/*
 *  linux/drivers/char/pty.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  Added support for a Unix98-style ptmx device.
 *    -- C. Scott Ananian <cananian@alumni.princeton.edu>, 14-Jan-1998
 *  Added TTY_DO_WRITE_WAKEUP to enable n_tty to send POLL_OUT to
 *      waiting writers -- Sapan Bhatia <sapan@corewars.org>
 *
 *
 */


#include <linux/autoconf.h>
#include <linux/version.h>

#if ! defined (LINUX_VERSION_CODE)
#error "Kernel version is not set"
#endif

// For kernel 2.4 and lower, add symbol versioning if configured in the kernel.
// Do not include for kernel 2.6 as the kbuild process turns on this option
//	 automatically if required.

#if (LINUX_VERSION_CODE < 0x20500)
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#  define MODVERSIONS /* force it on */
#endif

#ifdef MODVERSIONS
#  include <linux/modversions.h>
#endif
#endif	//LINUX_VERSION_CODE


#include <linux/module.h>	/* For EXPORT_SYMBOL */

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/fcntl.h>
#include <linux/string.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/init.h>
//#include <linux/devfs_fs_kernel.h>
#if (LINUX_VERSION_CODE >= 0x20500)
#include <linux/sysctl.h>
#endif
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/bitops.h>

#define BUILDING_PTY_C 1
#include <linux/devpts_fs.h>

#include <linux/serial.h>

#include "../crd.h"
#include "ptyx.h"

// Include compat.h file for kernel version 2.2
#if (LINUX_VERSION_CODE < 0x20400)
#include "compat.h"
#endif


//unsigned long  ptyx_debug_level = -1;
unsigned long  ptyx_debug_level =	PTYX_DEBUG_OPEN 
											|  PTYX_DEBUG_CLOSE
											|  PTYX_DEBUG_WRITE
											|  PTYX_DEBUG_IOCTL
											|  PTYX_DEBUG_WAIT_UNTIL_SENT
											|  PTYX_DEBUG_BLOCK_TIL_READY
											|  PTYX_DEBUG_TERMIOS
											|  PTYX_DEBUG_MISC
//											|  PTYX_DEBUG_SPIN_LOCK
;

//unsigned long  ptyx_debug_level = 0;



/*
    The slave_active flag is necessary because if the open fails, and returns
    an error, the tty layer calls the close to cleanup.  Unfortunately, if
    the open failed because the network connection is down, when the close
    is called, it tries to "wait_til_sent" and delays for 30 seconds by default.
    The slave_active flag is used to bypass the wait_til_sent when there are no
    active slaves.
*/

/*
	on opens we have 2 modes of operation
	Wait for TCP connection
		In this mode, we will wait the specified time for the TCP connection to
		be established.  On return, if the TCP connection is down, we will return
		an error.
	Don't wait for TCP connection 
		In this mode, we do not wait for the TCP connection to be established.
		On return, if the TCP connection is down, we will return no error.
		
	In either case, blocking and clocal settings are obeyed.  So if we need
	to wait for DCD, we will wait for the network connection to be established.
*/

#if (LINUX_VERSION_CODE < 0x20500)


static struct tty_driver ptyx_master_driver, ptyx_slave_driver;

/* Note: set of tables for BSD */
static struct tty_struct *pty_table[PTYX_NUM_PTYS];
static struct termios *pty_termios[PTYX_NUM_PTYS];
static struct termios *pty_termios_locked[PTYX_NUM_PTYS];
static struct tty_struct *ttyp_table[PTYX_NUM_PTYS];
static struct termios *ttyp_termios[PTYX_NUM_PTYS];
static struct termios *ttyp_termios_locked[PTYX_NUM_PTYS];

#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif
#ifndef MAX
#define MAX(a,b)	((a) < (b) ? (b) : (a))
#endif
#define	TTY_DRIVER(x)	(tty->driver.x)

#else

static struct tty_driver *ptyx_master_driver, *ptyx_slave_driver;
#define TTY_DRIVER(x)	(tty->driver->x)

#endif	// LINUX_VERSION_CODE


static struct ptyx_struct pty_state[PTYX_NUM_PTYS];

static void set_xctrl_status(struct tty_struct * tty,unsigned int status)
{
	struct ptyx_struct *ptyx_info=(struct ptyx_struct *) tty->driver_data;

	tty->ctrl_status |= TIOCPKT_XCTRLSTATUS;
	ptyx_info->ctrl_status |= status;
}

static void clr_xctrl_status(struct tty_struct * tty,unsigned int status)
{
	struct ptyx_struct *ptyx_info=(struct ptyx_struct *) tty->driver_data;

	tty->ctrl_status |= TIOCPKT_XCTRLSTATUS;
	ptyx_info->ctrl_status &= (~status);
}

static void ptyx_close(struct tty_struct *tty, struct file * filp)
{
	struct ptyx_struct *ptyx_info;

#ifdef PTYX_DEBUG_OPEN
	int line;
#endif

	if (!tty)
	{
		return;
	}

#if (LINUX_VERSION_CODE < 0x20500)
	line = MINOR(tty->device) - TTY_DRIVER(minor_start);
#else
	line = tty->index - TTY_DRIVER(minor_start);
#endif
	ptyx_print(PTYX_DEBUG_OPEN, ("ptyx_close(%d): Enter-%s tty=0x%p, jiffies=%lu \n", line, 
		((TTY_DRIVER(subtype) == PTY_TYPE_MASTER) ? "MASTER": "SLAVE"), tty, jiffies));

	if (TTY_DRIVER(subtype) == PTY_TYPE_MASTER) 
	{
		if (tty->count > 1)
		{
			ptyx_print(PTYX_DEBUG_OPEN, ("pty_close(%d):MASTER count=%d!!\n", line, tty->count));
		}
	} 
	else 
	{
		if (tty->count > 2)
		{
 			ptyx_print(PTYX_DEBUG_OPEN, ("pty_close(%d): Exit-SLAVE count=%d!!\n", line, tty->count));
			return;
		}
	}

	ptyx_info = (struct ptyx_struct *) tty->driver_data;
	if (ptyx_info == NULL)
	{
 			ptyx_print(PTYX_DEBUG_OPEN, ("pty_close(%d): Exit-%s ptyx_info=tty->driver_data = NULL \n",line,
				 ((TTY_DRIVER(subtype) == PTY_TYPE_MASTER) ? "MASTER": "SLAVE") ));
	 	return;
	}

	wake_up_interruptible(&tty->read_wait);
	wake_up_interruptible(&tty->write_wait);
	tty->packet = 0;
	if (!tty->link)
	{
 			ptyx_print(PTYX_DEBUG_OPEN, ("pty_close(%d): Exit-%s tty->link = NULL \n",line,
				 ((TTY_DRIVER(subtype) == PTY_TYPE_MASTER) ? "MASTER": "SLAVE") ));
		return;
	}
	//tty->link->packet = 0;
	wake_up_interruptible(&tty->link->read_wait);
	wake_up_interruptible(&tty->link->write_wait);
	if (TTY_DRIVER(subtype) == PTY_TYPE_MASTER) 
	{
		set_bit(TTY_OTHER_CLOSED, &tty->flags);
		tty_hangup(tty->link);
		ptyx_info->delta_msr_wait_done = TRUE;
		wake_up_interruptible(&ptyx_info->delta_msr_wait);
		wake_up_interruptible(&ptyx_info->open_wait);
		ptyx_info->pending_drain_end=0;
		ptyx_info->network_status=0;
	}
	// if this is the last close then inform the daemon
	else 
	{
		if (tty->link && tty->link->packet /*&& (tty->count == 1)*/)
		{
			if (ptyx_info->closing_wait != USF_CLOSING_WAIT_NONE && ptyx_info->slave_active)
			{
				tty_wait_until_sent(tty, msec_to_jiffies((long)ptyx_info->closing_wait));
			}
			// drop signals if required
			if (tty->termios->c_cflag & HUPCL)
			{
				ptyx_info->modem_status &= ~(TIOCM_DTR|TIOCM_RTS);
				set_xctrl_status(tty,XCTRLSTATUS_SIGNALS);  // update server signals (DTR,RTS)
				wake_up_interruptible(&tty->link->read_wait);
			}
			
			clr_xctrl_status(tty,XCTRLSTATUS_OPEN);
			set_xctrl_status(tty,XCTRLSTATUS_CLOSE);
			wake_up_interruptible(&tty->link->read_wait);
			if (ptyx_info->blocked_open)
			{
				if (ptyx_info->close_delay)
				{
					set_current_state(TASK_INTERRUPTIBLE);
					schedule_timeout(msec_to_jiffies((long)ptyx_info->close_delay));
				}
				wake_up_interruptible(&ptyx_info->open_wait);
			}
		}
		ptyx_info->slave_active=0;
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(msec_to_jiffies((long)500));
	}
	set_bit(TTY_OTHER_CLOSED, &tty->link->flags);

	ptyx_print(PTYX_DEBUG_OPEN, ("ptyx_close(%d): Exit-%s tty=0x%p, jiffies=%lu \n", line, 
		((TTY_DRIVER(subtype) == PTY_TYPE_MASTER) ? "MASTER": "SLAVE"), tty, jiffies));

}



/*
 * The unthrottle routine is called by the line discipline to signal
 * that it can receive more characters.  For PTY's, the TTY_THROTTLED
 * flag is always set, to force the line discipline to always call the
 * unthrottle routine when there are fewer than TTY_THRESHOLD_UNTHROTTLE 
 * characters in the queue.  This is necessary since each time this
 * happens, we need to wake up any sleeping processes that could be
 * (1) trying to send data to the pty, or (2) waiting in wait_until_sent()
 * for the pty buffer to be drained.
 */
static void ptyx_unthrottle(struct tty_struct * tty)
{
	struct tty_struct *o_tty = tty->link;

	if (!o_tty)
		return;
#if (LINUX_VERSION_CODE <  0x2061b)	// Less than 2.6.27 kernel.
	if ((o_tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && o_tty->ldisc.write_wakeup)
		(o_tty->ldisc.write_wakeup)(o_tty);
#else
	if ((o_tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && o_tty->ldisc.ops->write_wakeup)
		(o_tty->ldisc.ops->write_wakeup)(o_tty);
#endif
	wake_up_interruptible(&o_tty->write_wait);
	set_bit(TTY_THROTTLED, &tty->flags);
}

int dump_data(struct tty_struct *tty)
{
	struct ptyx_struct *ptyx_info=(struct ptyx_struct *) tty->driver_data;
	int val2ret=0;
	
	if (TTY_DRIVER(subtype) == PTY_TYPE_SLAVE && ptyx_info->network_status==0 &&
		ptyx_info->open_wait_time==-1)
	{
		val2ret=1;
	}
	return(val2ret);
}

/*
 * WSH 05/24/97: modified to 
 *   (1) use space in tty->flip instead of a shared temp buffer
 *	 The flip buffers aren't being used for a pty, so there's lots
 *	 of space available.  The buffer is protected by a per-pty
 *	 semaphore that should almost never come under contention.
 *   (2) avoid redundant copying for cases where count >> receive_room
 * N.B. Calls from user space may now return an error code instead of
 * a count.
 */
 
#if (LINUX_VERSION_CODE < 0x2060A)		// less then 2.6.10
 
static int ptyx_write(struct tty_struct * tty, int from_user,
		       const unsigned char *buf, int count)
{
	struct tty_struct	*to = tty->link;
	int					c = 0, n, room;
	char				*temp_buffer;
	struct ptyx_struct *ptyx_info=(struct ptyx_struct *) tty->driver_data;

	if (!to || tty->stopped)
		return 0;

	if (from_user) 
	{
		down(&tty->flip.pty_sem);
		temp_buffer = &tty->flip.char_buf[0];
		while (count > 0) 
		{
			/* check space so we don't copy needlessly */ 
			n = to->ldisc.receive_room(to);
			if (n > count)
				n = count;
			if (!n) break;

#if (LINUX_VERSION_CODE < 0x20500)
			n  = MIN(n, PTY_BUF_SIZE);
#else
			n  = min(n, PTY_BUF_SIZE);
#endif

			n -= copy_from_user(temp_buffer, buf, n);
			if (!n) 
			{
				if (!c)
					c = -EFAULT;
				break;
			}

			/* check again in case the buffer filled up */
			room = to->ldisc.receive_room(to);
			if (n > room)
				n = room;
			if (!n) break;
			buf   += n; 
			c     += n;
			count -= n;
			if (!dump_data(tty))
			{
				to->ldisc.receive_buf(to, temp_buffer, 0, n);
			}
			else
			{
				ptyx_print(PTYX_DEBUG_WRITE, ("%s: dumping data from user %d\n",__FUNCTION__,n));
			}
		}
		up(&tty->flip.pty_sem);
	} else {
		c = to->ldisc.receive_room(to);
		if (c > count)
			c = count;
        
		if (!dump_data(tty))
		{
#if (LINUX_VERSION_CODE <  0x2061b)	// Less than 2.6.27 kernel.
			to->ldisc.receive_buf(to, buf, 0, c);
#else
			to->ldisc.ops->receive_buf(to, buf, 0, c);
#endif
		}
		else
		{
			ptyx_print(PTYX_DEBUG_WRITE, ("%s: dumping data not from user %d\n",__FUNCTION__,c));
		}
	}
	if (TTY_DRIVER(subtype) == PTY_TYPE_SLAVE)
	{
		ptyx_info->icount.tx+=c;
	}
	else
	{
		ptyx_info->icount.rx+=c;
	}
	
	return c;
}

#else		// linux kernel is at least 2.6.10 - write prototype has "from_user"
			//	 removed from parameter list.

static int ptyx_write(struct tty_struct * tty, 
				       const unsigned char *buf, int count)
{
	struct tty_struct	*to = tty->link;
	int					c;
	struct ptyx_struct *ptyx_info=(struct ptyx_struct *) tty->driver_data;

	if (!to || tty->stopped)
		return 0;

#if (LINUX_VERSION_CODE < 0x20610)	/* Less than 2.6.16 kernel */
	c = to->ldisc.receive_room(to);
#else
	c = to->receive_room;
#endif
	if (c > count)
		c = count;
	
	if (!dump_data(tty))
	{
#if (LINUX_VERSION_CODE <  0x2061b)	// Less than 2.6.27 kernel.
		to->ldisc.receive_buf(to, buf, 0, c);
#else
		to->ldisc.ops->receive_buf(to, buf, 0, c);
#endif
	}
	if (TTY_DRIVER(subtype) == PTY_TYPE_SLAVE)
	{
		ptyx_info->icount.tx+=c;
	}
	else
	{
		ptyx_info->icount.rx+=c;
	}
	return c;
}
#endif		// > 2.6.10



static int ptyx_write_room(struct tty_struct *tty)
{
	struct tty_struct *to = tty->link;

	if (!to || tty->stopped)
		return 0;

#if (LINUX_VERSION_CODE < 0x20610)	/* Less than 2.6.16 kernel */
	return to->ldisc.receive_room(to);
#else
	return to->receive_room;
#endif
}

/*
 *	WSH 05/24/97:  Modified for asymmetric MASTER/SLAVE behavior
 *	The chars_in_buffer() value is used by the ldisc select() function 
 *	to hold off writing when chars_in_buffer > WAKEUP_CHARS (== 256).
 *	The pty driver chars_in_buffer() Master/Slave must behave differently:
 *
 *      The Master side needs to allow typed-ahead commands to accumulate
 *      while being canonicalized, so we report "our buffer" as empty until
 *	some threshold is reached, and then report the count. (Any count >
 *	WAKEUP_CHARS is regarded by select() as "full".)  To avoid deadlock 
 *	the count returned must be 0 if no canonical data is available to be 
 *	read. (The N_TTY ldisc.chars_in_buffer now knows this.)
 *  
 *	The Slave side passes all characters in raw mode to the Master side's
 *	buffer where they can be read immediately, so in this case we can
 *	return the true count in the buffer.
 */
static int ptyx_chars_in_buffer(struct tty_struct *tty)
{
	struct tty_struct *to = tty->link;
	int count;

#if (LINUX_VERSION_CODE <  0x2061b)	// Less than 2.6.27 kernel.
	if (!to || !to->ldisc.chars_in_buffer)
		 return 0;

	/* The ldisc must report 0 if no characters available to be read */
	count = to->ldisc.chars_in_buffer(to);
#else
	if (!to || !to->ldisc.ops->chars_in_buffer)
		 return 0;

	/* The ldisc must report 0 if no characters available to be read */
	count = to->ldisc.ops->chars_in_buffer(to);
#endif

	if (TTY_DRIVER(subtype) == PTY_TYPE_SLAVE) 
		return count;

	/* Master side driver ... if the other side's read buffer is less than 
	 * half full, return 0 to allow writers to proceed; otherwise return
	 * the count.  This leaves a comfortable margin to avoid overflow, 
	 * and still allows half a buffer's worth of typed-ahead commands.
	 */
	return ((count < N_TTY_BUF_SIZE/2) ? 0 : count);
}


/* Set the lock flag on a pty */
static int ptyx_set_lock(struct tty_struct *tty, int * arg)
{
	int val;
	if (get_user(val, arg))
		return -EFAULT;
	if (val)
		set_bit(TTY_PTY_LOCK, &tty->flags);
	else
		clear_bit(TTY_PTY_LOCK, &tty->flags);
	return 0;
}

static int get_serial_info(struct ptyx_struct *ptyx_info,
			   struct serial_struct * retinfo)
{
	struct serial_struct tmp;
   
	if (!retinfo)
		return -EFAULT;
    
	memset(&tmp, 0, sizeof(tmp));
	tmp.close_delay = ptyx_info->close_delay/10;
	tmp.closing_wait =  (ptyx_info->closing_wait==USF_CLOSING_WAIT_NONE)?ASYNC_CLOSING_WAIT_NONE:ptyx_info->closing_wait/10;
	if (copy_to_user(retinfo,&tmp,sizeof(*retinfo)))
		return -EFAULT;
	return 0;
}

static int set_serial_info(struct ptyx_struct *ptyx_info, struct serial_struct * new_info)
{
	struct serial_struct new_serial;
	int  retval = 0;
	int close_delay,closing_wait;

	if (copy_from_user(&new_serial,new_info,sizeof(new_serial)))
	{
		return -EFAULT;
	}
	close_delay=new_serial.close_delay*10;
	closing_wait=(new_serial.closing_wait==ASYNC_CLOSING_WAIT_NONE)?USF_CLOSING_WAIT_NONE:new_serial.closing_wait*10;
	if (!capable(CAP_SYS_ADMIN)) {
		if (close_delay != ptyx_info->close_delay)
		{
			return -EPERM;
		}
		goto check_and_exit;
	}

	/*
	 * OK, past this point, all the error checking has been done.
	 * At this point, we start making changes.....
	 */

	ptyx_info->close_delay = close_delay;
	ptyx_info->closing_wait = closing_wait;


check_and_exit:
	return retval;
}


// ptyx_slave_stop() and ptyx_slave_start()
// 
// These routines are called before setting or resetting tty->stopped.
// They enable or disable output, as necessary.
//
static void ptyx_slave_stop(struct tty_struct *tty)
{
	set_xctrl_status(tty,XCTRLSTATUS_STOP);
	wake_up_interruptible(&tty->link->read_wait);
	return;
}


static void ptyx_slave_start(struct tty_struct *tty)
{
	set_xctrl_status(tty,XCTRLSTATUS_START);
	wake_up_interruptible(&tty->link->read_wait);
	return;
}

//*****************************************************************************
//
//		slave ioctl functions
//
//	NOTE:	The TIOCMGET and TIOCMSET functions/ioctl commands rely on the
//			driver_data member of the slave and master ttys pointing to the
//			the same data (ie. same memory location).
//*****************************************************************************

static int ptyx_slave_ioctl(struct tty_struct *tty, struct file *file,
									unsigned int cmd, unsigned long arg)
{
	struct ptyx_struct * ptyx_info;

	struct tty_struct * real_tty;
	struct serial_icounter_struct cnow,cprev;
	unsigned long flags;
	int line=-1;			// init to invalid
	int retval;

	if (!tty) 
	{
		 ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(???): called with NULL tty!\n"));
		return -EIO;
	}

#if (LINUX_VERSION_CODE < 0x20500)
	line = MINOR(tty->device) - TTY_DRIVER(minor_start);
#else
	line = tty->index - TTY_DRIVER(minor_start);
#endif
	ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): called cmd=%x\n", line, cmd));
	
	if (TTY_DRIVER(type) == TTY_DRIVER_TYPE_PTY &&
	    TTY_DRIVER(subtype) == PTY_TYPE_MASTER)
		real_tty = tty->link;
	else
		real_tty = tty;
	
	ptyx_info = (struct ptyx_struct *)real_tty->driver_data;
	
#ifdef PTYX_PARANOIA_CHECK
	// check if our private data pointer is NULL
	if ( ptyx_info == NULL)
	{

		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): called with tty->driver_data=ptyx_info = NULL \n", line));
		return -EIO; 
	}
	if (tty->link && tty->link->driver_data == NULL)
	{
		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): called with tty->link->driver_data=ptyx_info = NULL \n", line));
 		return -EIO; 
	}
#endif // PTYX_PARANOIA_CHECK
	
	switch(cmd) 
	{
	
	case TIOCSBRK:
	case TCSBRK:
	case TCSBRKP:
	{
		if (tty->link && tty->link->packet) 
		{
			set_xctrl_status(tty,XCTRLSTATUS_BREAK);
			wake_up_interruptible(&tty->link->read_wait);
			schedule();
		}
		return 0;
	}
#if (LINUX_VERSION_CODE < 0x20500)
	case TIOCMBIC:
	case TIOCMBIS:
	case TIOCMSET: 
	{
		unsigned int new_val;

		if (get_user(arg, (unsigned int *) arg))
			return -EFAULT;

		new_val = ptyx_info->modem_status;
		switch (cmd) 
		{
			case TIOCMBIC:
				new_val &= ~arg;
				break;
			case TIOCMBIS:
				new_val |=  arg;
				break;
			case TIOCMSET: 
				new_val = arg;
				break;
		}

		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): TIOCMBIC/BIS/SET:\n    arg=%0lx, nval=%0x, oval=%0x.\n",
					line,arg, new_val, ((struct ptyx_struct*)real_tty->driver_data)->modem_status ));
                    
        ptyx_info->modem_status = new_val;
    
        if (real_tty->link && real_tty->link->packet) 
        {
            set_xctrl_status(real_tty,XCTRLSTATUS_SIGNALS);
            wake_up_interruptible(&real_tty->link->read_wait);
            schedule();
        }

		return 0;
	}
	case TIOCMGET:
	{
		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): TIOCMGET: value = %0x.\n", line,
					((struct ptyx_struct*)real_tty->driver_data)->modem_status ));

		if (put_user (ptyx_info->modem_status, (unsigned int*)arg))
			return -EFAULT;
        
		return 0;
	}
#endif
	case TIOCGICOUNT:
	{
		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): cmd=TIOCGICOUNT \n", line));
		cnow = ptyx_info->icount;
		
		if (copy_to_user((void *)arg, &cnow, sizeof(cnow)))
			return -EFAULT;
		
		return 0;
	}
		
	case TIOCMIWAIT:
	{
		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): cmd=TIOCMIWAIT \n", line));
		/* get the current counters */
		cprev = ptyx_info->icount;
		while (1)
		{
			PTYX_LOCK(&ptyx_info->port_lock, flags);
			ptyx_info->delta_msr_wait_done = FALSE;
			PTYX_UNLOCK(&ptyx_info->port_lock, flags);
			wait_event_interruptible( ptyx_info->delta_msr_wait, (ptyx_info->delta_msr_wait_done == TRUE ) );
			/* see if a signal did it */
			if (signal_pending(current))
				return -ERESTARTSYS;
			
			cnow = ptyx_info->icount;
			if (cnow.rng == cprev.rng && cnow.dsr == cprev.dsr &&
					cnow.dcd == cprev.dcd && cnow.cts == cprev.cts)
			{
				return -EIO; /* no change => error */
			}
			if ( ((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
					((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
					((arg & TIOCM_CD)  && (cnow.dcd != cprev.dcd)) ||
					((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)) )
			{
				return 0;
			}
			cprev = cnow;
		}
	
		return 0;
	}	
	case TIOCGSERIAL:
	{
		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): TIOCGSERIAL:  arg=%0lx.\n", line, arg));

		return get_serial_info(ptyx_info, (struct serial_struct *) arg);
	}				   
	case TIOCSSERIAL:
	{
		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): TIOCSSERIAL:  arg=%0lx.\n", line, arg));
		return set_serial_info(ptyx_info,
					   (struct serial_struct *) arg);
		break;
	} 
	case TCFLSH:
	{
		// The linux tty driver doesn't have a flush input routine for 
		// the driver, assuming all backed up data is in the line discipline
		// buffers.  However, this is deffinately not true in our case
		// and for that mater any hardware that has recieve FIFO's.  
		// Here, we service the ioctl, but then lie and say we didn't
		// so the line discipline will process the flush also to remove
		// any buffers from the line disipline
		
		if ( (retval = tty_check_change(real_tty)) )
			return retval;

		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): TCFLSH:  arg=%0lx.\n", line, arg));

		switch(arg) 
		{
		case TCIFLUSH:
		case TCIOFLUSH:
		{
		 	ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): TCFLSH: doing TCIFLUSH/TCIOFLUSH \n", line));
			set_xctrl_status(real_tty,XCTRLSTATUS_FLUSHREAD);
			wake_up_interruptible(&tty->link->read_wait);

			if (arg == TCIFLUSH) 
				 break;
		}
		case TCOFLUSH: /* flush output, or the receive buffer */
		{	
 		 	ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): TCFLSH: doing TCOFLUSH \n", line));
			// This is handled in the tty_ioctl.c code with will call my ptyx_slave_flush_buffer
			break;
		}
		default:
		{
			/* POSIX.1 says return EINVAL if we got a bad arg */
			return (-EINVAL);
		}
		} // end switch
		/* pretend we didn't recognize this IOCTL */
		return(-ENOIOCTLCMD);
	}
	//	TCSETA,TCSETS,TCSETAW,TCSETSW handled by tty layer and our set_termios
	// functions
	case TCSETAF:
	case TCSETSF:
	{
		// The linux tty driver doesn't have a flush input routine for 
		// the driver, assuming all backed up data is in the line discipline
		// buffers.  However, this is deffinately not true in our case
		// an for that mater any hardware that has recieve FIFO's.  
		// Here, we service the ioctl, but then lie and say we didn't
		// so the line discipline will process the flush also to remove
		// any buffers from the line disipline

		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): TCSETAF/TCSETSF \n", line));

		if ( (retval = tty_check_change(real_tty)) )
			return retval;
	  
		set_xctrl_status(real_tty,XCTRLSTATUS_FLUSHREAD);
		wake_up_interruptible(&tty->link->read_wait);

		// pretend we didn't recognize this, so tty will do the wait to drain and
		// flush the line displine recieve buffers
		return(-ENOIOCTLCMD);
	}
	case TCXONC:
	{
		// The Linux Line Discipline (LD) handles TCXONC for us if we let 
		// it, but for TCION/TCIOFF it sends down the XOFF and XON characters, wether SW 
		// flow control is on or not.  We will handle the  TCXONC 
		// ioctl so that we can  do this the "right way" based on the 
		// setting of hardware or software flow control.
		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): TCXONC:  arg=%0lx.\n", line, arg));

		if ( (retval = tty_check_change(real_tty)) )
			return retval;

		switch(arg) 
		{
		case TCOON:
		{
			ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): TCXONC: doing TCOON \n", line ));
			ptyx_slave_start(real_tty);
			return 0;
		}
		case TCOOFF:
		{
			ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): TCXONC: doing TCOON \n", line ));
			ptyx_slave_stop(real_tty);
			return 0;
		}
		case TCION:
 		case TCIOFF:
		{
			unsigned int modem_status, rts_delta;
			ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): TCXONC: doing TCION \n", line ));

			// if RST flow control set then lower/raise RTS to flow off
			if ( _C_FLAG(real_tty, CRTSCTS) )
			{
	 			ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): TCXONC: CRTSCTS set \n", line ));
				PTYX_LOCK(&ptyx_info->port_lock, flags);
				modem_status = ptyx_info->modem_status;
				// if RTS up then bring it down to HW flow off and inform dserver 
				if ( modem_status & TIOCM_RTS	)
				{
					if (arg == TCIOFF)
					{
						 modem_status &= ~TIOCM_RTS;			// HW flow off		 
					}
				}
				else
				{
					if (arg == TCION)
					{
						modem_status |= TIOCM_RTS;
					}
				}
				// if RTS changed then tell daemon about it
				rts_delta = (modem_status ^ ptyx_info->modem_status) & TIOCM_RTS;
				ptyx_info->modem_status = modem_status;
				PTYX_UNLOCK(&ptyx_info->port_lock, flags);
				if (rts_delta)
				{
//					set_ctrl_status(ptyx_info,CTRLSTATUS_SIGNALS);
//			 		ctrl_status_snd(ptyx_info);		// inform daemon to get new ctrl status
					set_xctrl_status(real_tty,XCTRLSTATUS_SIGNALS);
					wake_up_interruptible(&tty->link->read_wait);

				}
			}
			// if SW flow control on then break and let tty handle sending XON or XOFF char
			if ( _I_FLAG(tty, IXOFF) )
			{
	 			ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_slave_ioctl(%d): TCXONC: IXOFF set \n", line ));
				break;
			}
			else
			{		// nothing to do
				 return 0;
			}
		}
		default:
		{
			return(-EINVAL);
		}
		}  // end switch
		// pretend we didn't recognize this, so tty will do it's normal processing
		return -ENOIOCTLCMD;
	}
	} // end switch

	return -ENOIOCTLCMD;
}

#if (LINUX_VERSION_CODE <  0x2061b)	// Kernel is less than 2.6.27
static void ptyx_slave_break_ctl(struct tty_struct *tty, int break_state)
{
    
	if (TTY_DRIVER(type) == TTY_DRIVER_TYPE_PTY &&
	    TTY_DRIVER(subtype) == PTY_TYPE_MASTER)
    {
    }
	if (break_state == -1)	// turn on break
	{
		set_xctrl_status(tty,XCTRLSTATUS_BREAKON);
	}
	else					// turn off break
	{
		set_xctrl_status(tty,XCTRLSTATUS_BREAKOFF);
	}
	wake_up_interruptible(&tty->link->read_wait);
//	schedule();
}
#else
static int ptyx_slave_break_ctl(struct tty_struct *tty, int break_state)
{
    
	if (TTY_DRIVER(type) == TTY_DRIVER_TYPE_PTY &&
	    TTY_DRIVER(subtype) == PTY_TYPE_MASTER)
    {
    }
	if (break_state == -1)	// turn on break
	{
		set_xctrl_status(tty,XCTRLSTATUS_BREAKON);
	}
	else					// turn off break
	{
		set_xctrl_status(tty,XCTRLSTATUS_BREAKOFF);
	}
	wake_up_interruptible(&tty->link->read_wait);
//	schedule();
	return 0;

}
#endif

#if (LINUX_VERSION_CODE >= 0x20500)
static int ptyx_slave_tiocmget(struct tty_struct *tty, struct file *file)
{
	struct tty_struct * real_tty;

	if (TTY_DRIVER(type) == TTY_DRIVER_TYPE_PTY &&
		    TTY_DRIVER(subtype) == PTY_TYPE_MASTER)
		real_tty = tty->link;
	else
		real_tty = tty;

	ptyx_print(PTYX_DEBUG_IOCTL, ("%s: value = %0x\n",__FUNCTION__,
				((struct ptyx_struct*)real_tty->driver_data)->modem_status ));

	return ((struct ptyx_struct*)real_tty->driver_data)->modem_status;
}


static int ptyx_slave_tiocmset(struct tty_struct *tty, struct file *file,
									       unsigned int set, unsigned int clear)
{
	unsigned int new_val;
	struct tty_struct * real_tty;
	struct ptyx_struct * ptyx_info;

	if (TTY_DRIVER(type) == TTY_DRIVER_TYPE_PTY &&
		    TTY_DRIVER(subtype) == PTY_TYPE_MASTER)
		real_tty = tty->link;
	else
		real_tty = tty;

	ptyx_info = (struct ptyx_struct *)real_tty->driver_data;
	new_val = ptyx_info->modem_status;
   	new_val	|= set;
	new_val &= ~clear;
   
	ptyx_print(PTYX_DEBUG_IOCTL, ("%s:\n   set=%0x; clear=%0x, nval=%0x, oval=%0x\n",__FUNCTION__,
				set, clear, new_val, ((struct ptyx_struct*)real_tty->driver_data)->modem_status ));


    ((struct ptyx_struct*)real_tty->driver_data)->modem_status = new_val;

    if (real_tty->link && real_tty->link->packet) 
    {
        set_xctrl_status(real_tty,XCTRLSTATUS_SIGNALS);
        wake_up_interruptible(&real_tty->link->read_wait);
        schedule();
    }
	return 0;
}
#endif


//*****************************************************************************
//
//		master ioctl functions
//
//	NOTE:	The TIOCMGET and TIOCMSET functions/ioctl commands rely on the
//			driver_data member of the slave and master ttys pointing to the
//			the same data (ie. same memory location).
//*****************************************************************************

static int ptyx_master_ioctl(struct tty_struct *tty, struct file *file,
									unsigned int cmd, unsigned long arg)
{
	struct tty_struct * real_tty;
	struct ptyx_struct *ptyx_info;
	unsigned int curdelta;
	int line;

	if (!tty) 
	{
		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_master_ioctl(): called with NULL tty!\n"));
		return -EIO;
	}

#if (LINUX_VERSION_CODE < 0x20500)
	line = MINOR(tty->device) - TTY_DRIVER(minor_start);
#else
	line = tty->index - TTY_DRIVER(minor_start);
#endif
		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_master_ioctl(%d): called cmd=%x\n", line,cmd));



	if (TTY_DRIVER(type) == TTY_DRIVER_TYPE_PTY &&
	    TTY_DRIVER(subtype) == PTY_TYPE_MASTER)
	{
		real_tty = tty->link;
	}
	else
	{
		real_tty = tty;
	}

	ptyx_info = (struct ptyx_struct *)tty->driver_data;
#ifdef PTYX_PARANOIA_CHECK
	// check if our private data pointer is NULL
	if ( ptyx_info == NULL)
	{
		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_master_ioctl(%d): called with tty->driver_data=ptyx_info = NULL \n", line));
		return -EIO; 
	}
	if (tty->link && tty->link->driver_data == NULL)
	{
		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_master_ioctl(%d): called with tty->link->driver_data=ptyx_info = NULL \n", line));
 		return -EIO; 
	}
#endif // PTYX_PARANOIA_CHECK

	switch(cmd) 
	{
	case TIOCSBRK:
	case TCSBRK:
	case TCSBRKP:
		tty_insert_flip_char(real_tty, 0, TTY_BREAK);
		tty_flip_buffer_push(real_tty);
		return 0;

	case TIOCSPTLCK: /* Set PT Lock (disallow slave open) */
		return ptyx_set_lock(tty, (int *) arg);
	case TCGETS:
		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_master_ioctl(%d): cmd=TCGETS\n", line));
#if (LINUX_VERSION_CODE >  0x20617)		// 2.6.23
		if (kernel_termios_to_user_termios((struct termios2 *)arg, real_tty->termios))
#else
		if (kernel_termios_to_user_termios((struct termios *)arg, real_tty->termios))
#endif
		{
			ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_master_ioctl(%d): return -EFAULT\n", line));

			return -EFAULT;
		}
		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_master_ioctl(%d): return 0\n", line));
		return 0;

	//this was added/modifed to work around the change made in kernel 2.6. 
	// The 2.6 tiocmset() function masks off all bits that are not 
	// DTE output signals which does not allow the COMredirect deamon to set 
	// the DCE ouput pin status in the driver.
		
	case PTX_IOCMSET:		
		if (get_user(arg, (unsigned int *) arg))
			return -EFAULT;

		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_master_ioctl(%d): PTX_IOCMSET: arg=%0lx, old value = %0x\n", line,
				arg, ((struct ptyx_struct*)tty->driver_data)->modem_status ));
		
		curdelta=(arg ^ ptyx_info->modem_status);
// only increment RI count on the trailing edge transition
		if ((curdelta & TIOCM_RI) && (arg & TIOCM_RI))
		{
			curdelta = curdelta & ~TIOCM_RI;
		}
		ptyx_info->modem_status = (ptyx_info->modem_status & (TIOCM_DTR|TIOCM_RTS)) | (arg & ~(TIOCM_DTR|TIOCM_RTS));
		if (curdelta & TIOCM_RI)
			ptyx_info->icount.rng++;
		
		if (curdelta & TIOCM_CD)
			ptyx_info->icount.dcd++;
		
		if (curdelta & TIOCM_DSR)
			ptyx_info->icount.dsr++;
		
		if (curdelta & TIOCM_CTS)
			ptyx_info->icount.cts++;
		
		if (curdelta)
		{
	 		ptyx_info->delta_msr_wait_done = TRUE;
			wake_up_interruptible(&ptyx_info->delta_msr_wait);
		}
		if (ptyx_info->modem_status & TIOCM_CD)
		{
			wake_up_interruptible(&ptyx_info->open_wait);
		}
		return 0;

	case PTX_IOCLSET:
		if (get_user(arg, (unsigned int *) arg))
			return -EFAULT;

		if (arg & LSRNODATA_OE)
			ptyx_info->icount.overrun++;

		if (arg & LSRNODATA_PE)
			ptyx_info->icount.parity++;
		
		if (arg & LSRNODATA_FE)
			ptyx_info->icount.frame++;
		
		if (arg & LSRNODATA_BI)
		{
			ptyx_info->icount.brk++;
			tty_insert_flip_char(real_tty, 0, TTY_BREAK);
			tty_flip_buffer_push(real_tty);
		}
		return 0;

	case PTX_IOCXCSGET:
		if (put_user (ptyx_info->ctrl_status, (unsigned int*)arg))
				return -EFAULT;
		
		ptyx_info->ctrl_status=0;
		return 0;

	case PTX_IOCDRAINEND:
		
		ptyx_info->pending_drain_end=0;
		return 0;
		
	case PTX_IOCNETSTAT:
		ptyx_info->network_status=arg;
        if (arg)
        {
            if (tty->link && ptyx_info->slave_active)
            {
                set_xctrl_status(tty->link,XCTRLSTATUS_SIGNALS);
                wake_up_interruptible(&tty->read_wait);
            }
        }
		wake_up_interruptible(&ptyx_info->open_wait);

		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_master_ioctl(%d): PTX_IOCNETSTAT network %s\n",line, ( arg ? "UP" : "DOWN")));

		return 0;
		
	case PTX_IOCOWTSET:
	{
		int wait_time;
		
		if (get_user(wait_time, (int *) arg))
			return -EFAULT;

		ptyx_info->open_wait_time=wait_time;
		return 0;
	}
#if (LINUX_VERSION_CODE < 0x20500)
	case TIOCMGET:
		ptyx_print(PTYX_DEBUG_IOCTL, ("ptyx_master_ioctl(%d): TIOCMGET: value = %0x.\n",line,
					((struct ptyx_struct*)tty->driver_data)->modem_status));
		if (put_user (((struct ptyx_struct*)tty->driver_data)->modem_status, (unsigned int*)arg))
				return -EFAULT;
		return 0;
#endif
			
	}
	return -ENOIOCTLCMD;
}



#if (LINUX_VERSION_CODE >= 0x20500)
static int ptyx_master_tiocmget(struct tty_struct *tty, struct file *file)
{
	ptyx_print(PTYX_DEBUG_IOCTL, ("%s: value = %0x.\n",__FUNCTION__,
					((struct ptyx_struct*)tty->driver_data)->modem_status));


#ifdef PTYX_PARANOIA_CHECK
	if (!tty || !tty->driver_data)
		 return -EIO;
#endif
	 return ((struct ptyx_struct*)tty->driver_data)->modem_status;
}

static int ptyx_master_tiocmset(struct tty_struct *tty, struct file *file,
									       unsigned int set, unsigned int clear)
{
	ptyx_print(PTYX_DEBUG_IOCTL, ("%s:\n   set=%0x; clear=%0x, nval=%0x, oval=%0x \n",__FUNCTION__,
				set, clear, (set & ~clear), ((struct ptyx_struct*)tty->driver_data)->modem_status ));

#ifdef PTYX_PARANOIA_CHECK
	if (!tty || !tty->driver_data)
		 return -EIO;
#endif
	((struct ptyx_struct*)tty->driver_data)->modem_status = set & ~clear;
	return 0;
}
#endif



//*****************************************************************************
static void ptyx_flush_buffer(struct tty_struct *tty)
{
	struct tty_struct *to = tty->link;
#if (LINUX_VERSION_CODE >=  0x2061a)	// Greater than or equal to 2.6.26 kernel.
	unsigned long flags;
#endif

	if (!to)
		return;
#if (LINUX_VERSION_CODE <  0x2061b)	// Less than 2.6.27 kernel.
	if (to->ldisc.flush_buffer)
		to->ldisc.flush_buffer(to);
#else
	if (to->ldisc.ops->flush_buffer)
		to->ldisc.ops->flush_buffer(to);
#endif

	if (to->packet) {
#if (LINUX_VERSION_CODE >=  0x2061a)	// Greater than or equal to 2.6.26 kernel.
        spin_lock_irqsave(&tty->ctrl_lock, flags);
#endif
		tty->ctrl_status |= TIOCPKT_FLUSHWRITE;
		wake_up_interruptible(&to->read_wait);
#if (LINUX_VERSION_CODE >=  0x2061a)	// Greater than or equal to 2.6.26 kernel.
        spin_unlock_irqrestore(&tty->ctrl_lock, flags);
#endif
	}
}


static int block_til_ready(struct tty_struct *tty, struct file * filp)
{
	DECLARE_WAITQUEUE(wait, current);
	struct ptyx_struct *ptyx_info=(struct ptyx_struct *) tty->driver_data;
	int retval=0;
	int do_clocal = 0;
#if (LINUX_VERSION_CODE >= 0x020400)
	int try;
#endif

	if (tty->flags & (1 << TTY_IO_ERROR))
	{
		return -EIO;
	}
    ptyx_print(PTYX_DEBUG_BLOCK_TIL_READY, ("%s: open_wait_time=%d nonblock=%s\n",__FUNCTION__,ptyx_info->open_wait_time,filp->f_flags & O_NONBLOCK?"yes":"no"));

// don't block on open for 2.2 kernel
#if !(LINUX_VERSION_CODE < 0x20400)
	if ( (ptyx_info->open_wait_time >= 0) || ( ptyx_info->open_wait_time == -2) )
	{
        unsigned long orig_jiffies=jiffies;
        long timeout=-1;
        long delaytime = msec_to_jiffies(250);   // delay 1 sec between checks
        
		if (ptyx_info->open_wait_time >= 0)
		{
			timeout = msec_to_jiffies((long)ptyx_info->open_wait_time);
		}
		
		// we need to wait for the network connection to be established
		add_wait_queue(&ptyx_info->open_wait, &wait);
		ptyx_info->blocked_open++;
		for (try=0;;try++)
		{
			set_current_state(TASK_INTERRUPTIBLE);
			if (ptyx_info->network_status)
				break;
			
			if (signal_pending(current))
			{
				retval = -ERESTARTSYS;
				break;
			}
			if ( (ptyx_info->open_wait_time != OPENWAIT_FOREVER) && 
				 ( (ptyx_info->open_wait_time == 0) || 
					time_after(jiffies, orig_jiffies + timeout) ||
					test_bit(TTY_OTHER_CLOSED, &tty->link->flags) ) )
			{
				// don't wait for network connection
				retval = -EIO;
				break;
			}
			schedule_timeout(delaytime);
		}
		set_current_state(TASK_RUNNING);
		remove_wait_queue(&ptyx_info->open_wait, &wait);
		ptyx_info->blocked_open--;
	}
#endif // end of not 2.2 kernel

	if (ptyx_info->network_status)
	{
		set_xctrl_status(tty,XCTRLSTATUS_SIGNALS);  // update server signals (DTR,RTS)
		wake_up_interruptible(&tty->link->read_wait);
	}
	
	if ((filp->f_flags & O_NONBLOCK))
	{
		ptyx_print(PTYX_DEBUG_BLOCK_TIL_READY, ("%s: returning %d\n",__FUNCTION__,retval));
		return(retval);
	}
	if (tty->termios->c_cflag & CLOCAL)	// don't need DCD
	{
		do_clocal = 1;
	}
	else
	{
		retval=0;	// reset the return value to 0 for now
	}
	add_wait_queue(&ptyx_info->open_wait, &wait);
	ptyx_info->blocked_open++;
	while (1)
	{
		set_current_state(TASK_INTERRUPTIBLE);
		if (do_clocal || (ptyx_info->modem_status & TIOCM_CD))
			break;
		
		if (signal_pending(current))
		{
			retval = -ERESTARTSYS;
			break;
		}
		schedule();
	}
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&ptyx_info->open_wait, &wait);
	ptyx_info->blocked_open--;
	return(retval);
}

static int ptyx_open(struct tty_struct *tty, struct file * filp)
{
	int	retval = -ENODEV;
	int	line = -1;		// initialize to invalid
	struct	ptyx_struct *ptyx_info;

	
	if (!tty || !tty->link)
	{
		ptyx_print(PTYX_DEBUG_OPEN, ("ptyx_open(???): Enter-%s Null tty or tty->link, tty=0x%p, tty->link=0x%p \n",  
			 ((TTY_DRIVER(subtype) == PTY_TYPE_MASTER) ? "MASTER": "SLAVE"), 
			 tty, ( (tty != NULL) ? (void *)-1 : tty->link ) ));
		goto out;
	}
	
#if (LINUX_VERSION_CODE < 0x20500)
	line = MINOR(tty->device) - TTY_DRIVER(minor_start);
#else
	line = tty->index - TTY_DRIVER(minor_start);
#endif

	if ((line < 0) || (line >= PTYX_NUM_PTYS))
	{
		ptyx_print(PTYX_DEBUG_OPEN, ("ptyx_open(???): Enter-%s Invalid port, line = %d \n",  
			 ((TTY_DRIVER(subtype) == PTY_TYPE_MASTER) ? "MASTER": "SLAVE"), line ));
		goto out;
	}
	
	ptyx_print(PTYX_DEBUG_OPEN, ("ptyx_open(%d): Enter-%s tty=0x%p, jiffies=%lu \n", line, 
			((TTY_DRIVER(subtype) == PTY_TYPE_MASTER) ? "MASTER": "SLAVE"), tty, jiffies));


	retval = -EIO;
	if (test_bit(TTY_OTHER_CLOSED, &tty->flags))
	{
		ptyx_print(PTYX_DEBUG_OPEN, ("ptyx_open(%d): Enter-%s TTY_OTHER_CLOSED is set \n", line, 
		((TTY_DRIVER(subtype) == PTY_TYPE_MASTER) ? "MASTER": "SLAVE") ));
		goto out;
	}

	if (test_bit(TTY_PTY_LOCK, &tty->link->flags))
		goto out;

	if (tty->link->count != 1)
	{
		ptyx_print(PTYX_DEBUG_OPEN, ("ptyx_open(%d): %s count != 1 \n", line, 
		((TTY_DRIVER(subtype) == PTY_TYPE_MASTER) ? "MASTER": "SLAVE") ));
		goto out;
	}

	ptyx_info = (struct ptyx_struct *)(TTY_DRIVER(driver_state)) + line;

	// This code is to resolve a bug intoduced in 2.26.26 tty_io.c where 
	// the Slave tty-ops is set to the masters tty_driver->ops
#if (LINUX_VERSION_CODE >=  0x2061a)	// Greater than or equal to 2.6.26 kernel.
	tty->ops = TTY_DRIVER(ops);
#endif

	tty->driver_data = ptyx_info;
	tty->link->driver_data = ptyx_info;		// no harm setting pointer on other side 

	clear_bit(TTY_OTHER_CLOSED, &tty->link->flags);
	wake_up_interruptible(&ptyx_info->open_wait);
	set_bit(TTY_THROTTLED, &tty->flags);
	set_bit(TTY_DO_WRITE_WAKEUP, &tty->flags);

	// if first open of slave device then inform the daemon
	if (TTY_DRIVER(subtype) == PTY_TYPE_SLAVE) 
	{
		if (tty->termios->c_cflag & CBAUD && ptyx_info->slave_active==0)
		{
			ptyx_info->modem_status |= (TIOCM_DTR | TIOCM_RTS);
		}
		if (tty->link && tty->link->packet ) 
		{
			clr_xctrl_status(tty,XCTRLSTATUS_CLOSE);
			set_xctrl_status(tty,XCTRLSTATUS_OPEN);
			wake_up_interruptible(&tty->link->read_wait);
		}

		retval = block_til_ready(tty, filp);
		if (retval)
		{
			 ptyx_print(PTYX_DEBUG_OPEN, ("ptyx_open(%d): returning after block_til_ready with %d\n",line,retval));
			goto out;
		}
        ptyx_info->slave_active |= 1;
	}
	retval = 0;

out:
 	ptyx_print(PTYX_DEBUG_OPEN, ("ptyx_open(%d): Exit-%s tty=0x%p, returning %d, jiffies=%lu \n", line, 
		((TTY_DRIVER(subtype) == PTY_TYPE_MASTER) ? "MASTER": "SLAVE"), tty, retval, jiffies));

	return retval;
}


#if (LINUX_VERSION_CODE >  0x20612)		// 2.6.18
static void ptyx_set_termios(struct tty_struct *tty, struct ktermios *old_termios)
#else
static void ptyx_set_termios(struct tty_struct *tty, struct termios *old_termios)
#endif
{
	struct ptyx_struct *ptyx_info;
	
	ptyx_info = (struct ptyx_struct *)tty->driver_data;
	
	ptyx_print(PTYX_DEBUG_TERMIOS, ("%s: called current->state=%lx\n",__FUNCTION__,current->state));
	if (tty->link && tty->link->packet) 
	{
		set_xctrl_status(tty,XCTRLSTATUS_TERMIOS);
		wake_up_interruptible(&tty->link->read_wait);
		schedule();
	} else {
		tty->termios->c_cflag &= ~(CSIZE | PARENB);
		tty->termios->c_cflag |= (CS8 | CREAD);
	}
	ptyx_print(PTYX_DEBUG_TERMIOS, ("%s: return current->state=%lx\n",__FUNCTION__,current->state));
}

#if (LINUX_VERSION_CODE > 0x2061b)		// greater than 2.6.27 kernel
static int ptyx_install(struct tty_driver *driver, struct tty_struct *tty)
{
	struct tty_struct *o_tty;
	int idx = tty->index;
	int retval;

	o_tty = alloc_tty_struct();
	if (!o_tty)
	{
		return -ENOMEM;
	}
	if (!try_module_get(driver->other->owner)) 
	{
		/* This cannot in fact currently happen */
		free_tty_struct(o_tty);
		return -ENOMEM;
	}
	initialize_tty_struct(o_tty, driver->other, idx);

	/* We always use new tty termios data so we can do this
	   the easy way .. */
	retval = tty_init_termios(tty);
	if (retval)
	{
		goto free_mem_out;
	}

	retval = tty_init_termios(o_tty);
	if (retval) {
		tty_free_termios(tty);
		goto free_mem_out;
	}

	/*
	 * Everything allocated ... set up the o_tty structure.
	 */
	driver->other->ttys[idx] = o_tty;
	tty_driver_kref_get(driver->other);
	if (driver->subtype == PTY_TYPE_MASTER)
	{
		o_tty->count++;
	}
	
	/* Establish the links in both directions */
	tty->link   = o_tty;
	o_tty->link = tty;

	tty_driver_kref_get(driver);
	tty->count++;
	driver->ttys[idx] = tty;
	return 0;
	
free_mem_out:
	module_put(o_tty->driver->owner);
	free_tty_struct(o_tty);
	return -ENOMEM;
}
#endif

static void ptyx_wait_until_sent(struct tty_struct *tty, int timeout)
{
	volatile struct ptyx_struct *info = (struct ptyx_struct *)tty->driver_data;
	unsigned long orig_jiffies, char_time;

	if (dump_data(tty))
	{
		return;
	}
	
    info->pending_drain_end=1;
	set_xctrl_status(tty,XCTRLSTATUS_DRAIN);
	wake_up_interruptible(&tty->link->read_wait);
	orig_jiffies = jiffies;
	char_time = msec_to_jiffies(10);
	ptyx_print(PTYX_DEBUG_WAIT_UNTIL_SENT, ("%s: timeout=%d check=%lu jiff=%lu\n", __FUNCTION__,timeout, char_time, jiffies));

	/* We go through the loop at least once because we can't tell
	 * exactly when the last character exits the shifter.  There can
	 * be at least two characters waiting to be sent after the buffers
	 * are empty.
	 */
	do {
// uncomment if you really want to know how times we've gone through while loop
//		ptyx_print(PTYX_DEBUG_WAIT_UNTIL_SENT, ("%s: (jiff=%lu)...\n",__FUNCTION__,jiffies));

		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(char_time);
		if (signal_pending(current))
			break;
		if (timeout && time_after(jiffies, orig_jiffies + timeout))
			break;
		
	} while (info->pending_drain_end);
	current->state = TASK_RUNNING;
	ptyx_print(PTYX_DEBUG_WAIT_UNTIL_SENT, ("%s: (jiff=%lu)...done\n",__FUNCTION__,jiffies));
}

#if (LINUX_VERSION_CODE >= 0x20500)
static struct tty_operations ptyx_ops_master = 
{
#if (LINUX_VERSION_CODE > 0x2061b)		// greater than 2.6.27 kernel
	.install = ptyx_install,
#endif
	.open = ptyx_open,
	.close = ptyx_close,
	.write = ptyx_write,
	.write_room = ptyx_write_room,
	.flush_buffer = ptyx_flush_buffer,
	.chars_in_buffer = ptyx_chars_in_buffer,
	.unthrottle = ptyx_unthrottle,
	.ioctl = ptyx_master_ioctl,
	.set_termios = ptyx_set_termios,
	.tiocmget = ptyx_master_tiocmget,
	.tiocmset = ptyx_master_tiocmset,
};

static struct tty_operations ptyx_ops_slave = 
{
#if (LINUX_VERSION_CODE > 0x2061b)		// greater than 2.6.27 kernel
	.install = ptyx_install,
#endif
	.open = ptyx_open,
	.close = ptyx_close,
	.write = ptyx_write,
	.write_room = ptyx_write_room,
	.flush_buffer = ptyx_flush_buffer,
	.chars_in_buffer = ptyx_chars_in_buffer,
	.unthrottle = ptyx_unthrottle,
	.ioctl = ptyx_slave_ioctl,
	.set_termios = ptyx_set_termios,
	.break_ctl = ptyx_slave_break_ctl,
	.wait_until_sent = ptyx_wait_until_sent,
	.tiocmget = ptyx_slave_tiocmget,
	.tiocmset = ptyx_slave_tiocmset,
};
#endif



static int __init ptyx_init(void)
{
	int i;

#if (LINUX_VERSION_CODE < 0x20500)		// kernel is 2.4 or less.

	printk("BLACK BOX COMredirect Driver v%s\n", TP_VERSION);

	/* Traditional BSD devices */
	memset(&pty_state, 0, sizeof(pty_state));
	for (i = 0; i < PTYX_NUM_PTYS; i++)
	{
		init_waitqueue_head(&pty_state[i].open_wait);
		init_waitqueue_head(&pty_state[i].delta_msr_wait);
		pty_state[i].close_delay = DEFAULT_CLOSE_DELAY;
		pty_state[i].closing_wait = DEFAULT_CLOSING_WAIT_TIME;
		pty_state[i].modem_status = TIOCM_DTR | TIOCM_RTS;    // the default baud is 38400 so default DTR and RTS on
		pty_state[i].open_wait_time = DEFAULT_OPEN_WAIT_TIME * 1000;	// convert to ms
		spin_lock_init(&pty_state[i].port_lock);
	}

// Master Driver 
	memset(&ptyx_master_driver, 0, sizeof(struct tty_driver));
	ptyx_master_driver.magic = TTY_DRIVER_MAGIC;

	ptyx_master_driver.driver_name = "ptyx_master";
	ptyx_master_driver.name = "ptyx";
	ptyx_master_driver.major = PTYX_MASTER_MAJOR;
	ptyx_master_driver.minor_start = 0;
	ptyx_master_driver.num = PTYX_NUM_PTYS;
	ptyx_master_driver.type = TTY_DRIVER_TYPE_PTY;
	ptyx_master_driver.subtype = PTY_TYPE_MASTER;
	ptyx_master_driver.init_termios = tty_std_termios;
	ptyx_master_driver.init_termios.c_iflag = 0;
	ptyx_master_driver.init_termios.c_oflag = 0;
	ptyx_master_driver.init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;
	ptyx_master_driver.init_termios.c_lflag = 0;
	ptyx_master_driver.flags = TTY_DRIVER_RESET_TERMIOS | TTY_DRIVER_REAL_RAW;
	ptyx_master_driver.refcount = (int*)&(THIS_MODULE->uc.usecount);
	ptyx_master_driver.table = pty_table;
	ptyx_master_driver.termios = pty_termios;
	ptyx_master_driver.termios_locked = pty_termios_locked;
	ptyx_master_driver.driver_state = pty_state;
	ptyx_master_driver.other = &ptyx_slave_driver;
	ptyx_master_driver.ioctl = ptyx_master_ioctl;
	ptyx_master_driver.open = ptyx_open;
	ptyx_master_driver.close = ptyx_close;
	ptyx_master_driver.write = ptyx_write;
	ptyx_master_driver.write_room = ptyx_write_room;
	ptyx_master_driver.flush_buffer = ptyx_flush_buffer;
	ptyx_master_driver.chars_in_buffer = ptyx_chars_in_buffer;
	ptyx_master_driver.unthrottle = ptyx_unthrottle;
	ptyx_master_driver.set_termios = ptyx_set_termios;

// Slave driver
	ptyx_slave_driver = ptyx_master_driver;			// copy structure
	ptyx_slave_driver.driver_name = "ptyx_slave";
	ptyx_slave_driver.proc_entry = 0;
	ptyx_slave_driver.name = "ttyxp";
	ptyx_slave_driver.subtype = PTY_TYPE_SLAVE;
	ptyx_slave_driver.major = PTYX_SLAVE_MAJOR;
	ptyx_slave_driver.minor_start = 0;
	ptyx_slave_driver.init_termios = tty_std_termios;
	ptyx_slave_driver.init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;
	/* Slave ptys are registered when their corresponding master pty
	 * is opened, and unregistered when the pair is closed.
	 */
	ptyx_slave_driver.flags |= TTY_DRIVER_NO_DEVFS;
	ptyx_slave_driver.table = ttyp_table;
	ptyx_slave_driver.termios = ttyp_termios;
	ptyx_slave_driver.termios_locked = ttyp_termios_locked;
	ptyx_slave_driver.driver_state = pty_state;
	ptyx_slave_driver.other = &ptyx_master_driver;
	ptyx_slave_driver.ioctl = ptyx_slave_ioctl;
	ptyx_slave_driver.break_ctl = ptyx_slave_break_ctl;
	ptyx_slave_driver.wait_until_sent = &ptyx_wait_until_sent;

	if (tty_register_driver(&ptyx_slave_driver))
		panic("Couldn't register pty slave driver");
	if (tty_register_driver(&ptyx_master_driver))
		panic("Couldn't register pty master driver");

		
#else	//LINUX_VERSION_CODE is 2.5 or greater.
		
	printk("BLACK BOX COMredirect Driver v%s\n", TP_VERSION);

	/* Traditional BSD devices */
	ptyx_master_driver = alloc_tty_driver(PTYX_NUM_PTYS);
	if (!ptyx_master_driver)
		panic("Couldn't allocate ptyx master driver");

	ptyx_slave_driver = alloc_tty_driver(PTYX_NUM_PTYS);
	if (!ptyx_slave_driver)
		panic("Couldn't allocate ptyx slave driver");

	memset(&pty_state, 0, sizeof(pty_state));
		
	for (i = 0; i < PTYX_NUM_PTYS; i++)
	{
		init_waitqueue_head(&pty_state[i].open_wait);
		init_waitqueue_head(&pty_state[i].delta_msr_wait);
		pty_state[i].close_delay = DEFAULT_CLOSE_DELAY;
		pty_state[i].closing_wait = DEFAULT_CLOSING_WAIT_TIME;
		pty_state[i].modem_status = TIOCM_DTR | TIOCM_RTS;    // the default baud is 38400 so default DTR and RTS on
		pty_state[i].open_wait_time = DEFAULT_OPEN_WAIT_TIME * 1000; // convert to ms
		spin_lock_init(&pty_state[i].port_lock);

	}
	
// Master Driver 
	ptyx_master_driver->owner = THIS_MODULE;
	ptyx_master_driver->driver_name = "ptyx_master";
	ptyx_master_driver->name = "ptyx";
	ptyx_master_driver->major = PTYX_MASTER_MAJOR;
	ptyx_master_driver->minor_start = 0;
	ptyx_master_driver->type = TTY_DRIVER_TYPE_PTY;
	ptyx_master_driver->subtype = PTY_TYPE_MASTER;
	ptyx_master_driver->init_termios = tty_std_termios;
	ptyx_master_driver->init_termios.c_iflag = 0;
	ptyx_master_driver->init_termios.c_oflag = 0;
	ptyx_master_driver->init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;
	ptyx_master_driver->init_termios.c_lflag = 0;
	ptyx_master_driver->flags = TTY_DRIVER_RESET_TERMIOS | TTY_DRIVER_REAL_RAW;
	ptyx_master_driver->driver_state = pty_state;
	ptyx_master_driver->other = ptyx_slave_driver;
	tty_set_operations(ptyx_master_driver, &ptyx_ops_master);


// Slave driver
	ptyx_slave_driver->owner = THIS_MODULE;
	ptyx_slave_driver->driver_name = "ptyx_slave";
	ptyx_slave_driver->name = "ttyxp";
	ptyx_slave_driver->major = PTYX_SLAVE_MAJOR;
	ptyx_slave_driver->minor_start = 0;
	ptyx_slave_driver->type = TTY_DRIVER_TYPE_PTY;
	ptyx_slave_driver->subtype = PTY_TYPE_SLAVE;
	ptyx_slave_driver->init_termios = tty_std_termios;
	ptyx_slave_driver->init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;
	ptyx_slave_driver->flags = TTY_DRIVER_RESET_TERMIOS | TTY_DRIVER_REAL_RAW;
	ptyx_slave_driver->driver_state = pty_state;
	ptyx_slave_driver->other = ptyx_master_driver;
	tty_set_operations(ptyx_slave_driver, &ptyx_ops_slave);

	if (tty_register_driver(ptyx_slave_driver))
		panic("Couldn't register pty slave driver");
	if (tty_register_driver(ptyx_master_driver))
		panic("Couldn't register pty driver");

#endif	// LINUX_VERSION_CODE
	 
	return 0;
}

void __exit ptyx_finish (void) 
{
	ptyx_print(PTYX_DEBUG_MISC, ("ptyx_finish(): unregistering driver \n"));

#if (LINUX_VERSION_CODE < 0x020500)
	tty_unregister_driver(&ptyx_master_driver);
	tty_unregister_driver(&ptyx_slave_driver);
#else
	tty_unregister_driver(ptyx_master_driver);
	tty_unregister_driver(ptyx_slave_driver);
#endif
}

module_init(ptyx_init);
module_exit(ptyx_finish);


MODULE_DESCRIPTION("BLACK BOX COMredirect pty driver");
MODULE_AUTHOR("");

#if (LINUX_VERSION_CODE >= 0x020400)
MODULE_LICENSE("GPL");
#endif


